import pytest

from .hash_cleaned_str import clean_str, hash_str_to_int


def hash_cleaned_str(input_str: str) -> int:
    return hash_str_to_int(clean_str(input_str))


def test_hash_cleaned_str_should_ignore_punctuation():
    strings_should_hash_to_identical = ["this", "This"]

    punctuation = r"!().,:;/"
    strings_should_hash_to_identical += [
        f"{s}{punctuation}" for s in strings_should_hash_to_identical
    ]

    assert (
        len(
            {
                hash_str_to_int(clean_str(s))
                for s in strings_should_hash_to_identical
            }
        )
        == 1
    )


@pytest.mark.parametrize(
    ("input_str", "hash_identical_str"), [("<p>Test</p>", "Test"), ("å", "å")]
)
def test_hash_cleaned_str_should_remove_html_tags(
    input_str: str, hash_identical_str: str
):
    assert hash_cleaned_str(input_str) == hash_cleaned_str(hash_identical_str)


@pytest.mark.parametrize(
    ("input_str", "expected"),
    [
        ("Is <2, but >4", "is_2_but_4"),
        ("Q. ![](src). Question?", "q_src_question"),
        ("pre <img src='testsrc' /> post", "pre_testsrc_post"),
        ("\t", ""),
        ("\n", ""),
        ("""1. One\n\t2. Two""", "one_two"),
        ("""* One\n\t* Two""", "one_two"),
        ("""- One\n\t- Two""", "one_two"),
        ("[link1](blah) and [link2](blah)", "link1_and_link2"),
    ],
)
def test_str_cleaner(input_str: str, expected: str):
    assert clean_str(input_str) == expected



from collections.abc import Sequence
from dataclasses import dataclass

import genanki

from ...utils.hash_cleaned_str import hash_str_to_int
from .anki_prompt import AnkiPrompt


@dataclass(frozen=True)
class AnkiCloze(AnkiPrompt):
    base_deck: str
    tags: Sequence[str]
    text: str
    css: str

    @property
    def genanki_model(self) -> genanki.Model:
        model_name = "Ankdown Cloze with UUID"

        return genanki.Model(
            model_id=hash_str_to_int(model_name),
            name=model_name,
            fields=[
                {"name": "Text"},
                {"name": "Extra"},
                {"name": "Tags"},
                {"name": "UUID"},
            ],
            templates=[
                {
                    "name": "Ankdown Cloze Card with UUID",
                    "qfmt": r"{{{{cloze:Text}}}}\n<div class='extra'>{{{{Extra}}}}</div>\n{}",
                    "afmt": r"{{{{cloze:Text}}}}\n<div class='extra'>{{{{Extra}}}}</div>\n{}",
                }
            ],
            css=self.css,
            model_type=1,  # This is the model_type number for genanki, takes 0 for QA or 1 for cloze
        )

    def to_genanki_note(self) -> genanki.Note:
        return genanki.Note(
            guid=str(self.uuid),
            model=self.genanki_model,
            fields=[
                self.field_to_html(self.text),
                "",
                " ".join(self.tags),
                str(self.uuid),
            ],
            tags=self.tags,
        )



from collections.abc import Mapping

import pytest

from ..source.prompts.prompt_cloze import ClozeWithoutDoc
from ..source.prompts.prompt_qa import QAWithoutDoc
from .ankiconnect.anki_converter import AnkiPromptConverter
from .ankiconnect.ankiconnect_gateway import (
    AnkiField,
    ImportPackage,
    SpieAnkiconnectGateway,
    UpdateModel,
)
from .ankiconnect.test_ankiconnect_gateway import MockNoteInfo
from .destination import PushPrompts
from .destination_ankiconnect import AnkiConnectDestination


@pytest.mark.parametrize(
    ("fields"),
    [
        ({"Text": AnkiField(value="MockText", order=0)}),
        (
            {
                "Question": AnkiField(value="MockQuestion", order=0),
                "Answer": AnkiField(value="MockAnswer", order=1),
            }
        ),
        pytest.param(
            ({"UnknownField": AnkiField(value="MockText", order=0)}),
            marks=pytest.mark.xfail,
        ),
    ],
)
def test_ankiconnect_get_all_prompts(fields: Mapping[str, AnkiField]):
    dest = AnkiConnectDestination(
        gateway=SpieAnkiconnectGateway(
            note_infos=[MockNoteInfo(fields=fields)]
        ),
        prompt_converter=AnkiPromptConverter(
            base_deck="FakeDeck", card_css="FakeCSS"
        ),
    )
    prompts = dest.get_all_prompts()

    assert len(prompts) == 1


def test_ankiconnect_push_prompts():
    gateway = SpieAnkiconnectGateway()
    dest = AnkiConnectDestination(
        gateway=gateway,
        prompt_converter=AnkiPromptConverter(
            base_deck="FakeDeck", card_css="FakeCSS"
        ),
    )
    dest.update(
        [
            PushPrompts(
                prompts=[
                    QAWithoutDoc(
                        question="FakeQuestion",
                        answer="FakeAnswer",
                        add_tags=["anki/deck/FakeSubdeck/FakeSubSubdeck"],
                    ),
                    ClozeWithoutDoc(text="FakeText", add_tags=["FakeTag"]),
                ]
            )
        ]
    )

    expected_commands = [(ImportPackage, 1), (UpdateModel, 2)]
    import_package_command = next(
        c for c in gateway.executed_commands if isinstance(c, ImportPackage)
    )
    assert (
        import_package_command.package.decks[0].name  # type: ignore
        == "FakeDeck::FakeSubdeck::FakeSubSubdeck"
    )
    assert len(import_package_command.package.decks) == 2  # type: ignore

    for command in expected_commands:
        assert (
            len(
                [
                    c
                    for c in gateway.executed_commands
                    if isinstance(c, command[0])
                ]
            )
            == command[1]
        )



from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path

import pytest

from ..document import Document
from ..prompts.prompt_qa import QAPrompt
from .extractor_table import TableExtractor


@dataclass(frozen=True)
class TableExtractorExample:
    table_prompt: str  # What the example is testing
    expectation: Sequence[QAPrompt]  # Expected prompts


@pytest.mark.parametrize(
    ("example"),
    [
        TableExtractorExample(
            table_prompt="Descending // |Column one|? // |Column two|",
            expectation=[QAPrompt(question="11?", answer="22")],
        ),
        TableExtractorExample(
            table_prompt="Ascending // |Column one|? // |Column two|",
            expectation=[
                QAPrompt(question="21?", answer="12"),
                QAPrompt(question="31?", answer="22"),
            ],
        ),
        TableExtractorExample(
            table_prompt="Rowwise // |Column one|? // |Column two|",
            expectation=[
                QAPrompt(question="11?", answer="12"),
                QAPrompt(question="21?", answer="22"),
            ],
        ),
    ],
)
def test_table_extractor(example: TableExtractorExample):
    input_doc = Document(
        content=f"""
    Block one

| Column one | Column two |
| --- | --- |
| 11 | 12 |
| 21 | 22 |
| 31 |      |
{example.table_prompt}""",
        source_path=Path(__file__),
    )

    assert set(example.expectation) == set(
        TableExtractor().extract_prompts(input_doc)
    )



import logging
from collections.abc import Sequence
from typing import Protocol

from iterpy import Iter

from .document import Document
from .document_source import BaseDocumentSource
from .extractors.extractor import BasePromptExtractor
from .prompts.prompt import BasePrompt
from .prompts.prompt_from_doc import PromptFromDocMixin

log = logging.getLogger(__name__)


class BasePromptSource(Protocol):
    def get_prompts(self) -> Sequence[BasePrompt]:
        ...


class DocumentPromptSource(BasePromptSource):
    def __init__(
        self,
        document_ingester: BaseDocumentSource,
        prompt_extractors: Sequence[BasePromptExtractor],
    ):
        self._document_ingester = document_ingester
        self._prompt_extractors = prompt_extractors

    def _get_prompts_from_document(
        self, document: Document
    ) -> Sequence[PromptFromDocMixin]:
        prompts: list[PromptFromDocMixin] = []

        for extractor in self._prompt_extractors:
            extractor_prompts = list(extractor.extract_prompts(document))
            prompts += extractor_prompts  # type: ignore

        return prompts

    def _deduplicate_group(
        self, group: tuple[str, Sequence[PromptFromDocMixin]]
    ) -> BasePrompt:
        prompts_in_group = group[1]

        if len(prompts_in_group) != 1:
            duplicate_prompt_locations = (
                Iter(prompts_in_group)
                .map(
                    lambda prompt: f"{prompt.parent_doc.source_path.name}:{prompt.line_nr}"
                )
                .to_list()
            )
            log.warn(f"Found duplicate prompts in {duplicate_prompt_locations}")

        return prompts_in_group[0]

    def _deduplicate_prompts(
        self, prompts: Sequence[PromptFromDocMixin]
    ) -> Sequence[BasePrompt]:
        """Deduplicate prompts based on scheduling UID. If the scheduling UID is the same, the prompt is considered a duplicate."""
        scheduling_uuid_groups = Iter(prompts).groupby(
            lambda prompt: str(prompt.scheduling_uid)
        )

        unique_prompts = scheduling_uuid_groups.map(
            self._deduplicate_group
        ).to_list()

        n_duplicates = len(prompts) - len(unique_prompts)
        if n_duplicates != 0:
            log.warn(f"Found a total of {n_duplicates} duplicate prompts")

        return unique_prompts

    def get_prompts(self) -> Sequence[BasePrompt]:
        prompts = (
            Iter(self._document_ingester.get_documents())
            .map(self._get_prompts_from_document)
            .flatten()
            .to_list()
        )

        return self._deduplicate_prompts(prompts)



import logging
import re
from collections.abc import Sequence

from ..document import Document
from ..prompts.prompt_qa import QAFromDoc, QAPrompt
from .extractor import BasePromptExtractor

log = logging.getLogger(__name__)


class QAPromptExtractor(BasePromptExtractor):
    def __init__(self, question_prefix: str, answer_prefix: str) -> None:
        self.question_prefix = question_prefix
        self.answer_prefix = answer_prefix

    def _get_first_question(self, content: str) -> str:
        question = re.findall(
            self.question_prefix + r"{0,1}\.(?:(?!A\.).)*",
            content,
            flags=re.DOTALL,
        )[0]

        return question[len(self.question_prefix) + 1 :].rstrip()

    def _get_first_answer(self, content: str) -> str:
        # Have to use positive lookahead to match code-blocks
        # To ensure the last answer is matched as well, we add 2 newlines to string.
        string_padded = f"{content.rstrip()}\n\n"

        answer = re.findall(
            r"\n" + self.answer_prefix + r"[ \n]+\n*.+",
            string_padded,
            re.DOTALL,
        )[0]

        return answer[len(self.answer_prefix) + 2 :].rstrip()

    @staticmethod
    def _string_to_blocks_by_newlines(string: str) -> list[str]:
        """Break string into a list by 2+ newlines in a row."""
        return re.split(r"(\n\n)+", string)

    def _has_qa(self, string: str) -> bool:
        """Check whether a string contains a qa prompt"""
        return (
            len(
                re.findall(
                    r"^(?![:>]).*" + self.question_prefix + r"{0,1}\. ",
                    string,
                    flags=re.DOTALL,
                )
            )
            != 0
        )

    def extract_prompts(self, document: Document) -> Sequence[QAPrompt]:
        prompts: list[QAPrompt] = []
        blocks = self._string_to_blocks_by_newlines(document.content)
        block_starting_line_nr = 1

        for block_string in blocks:
            if self._has_qa(block_string):
                question = self._get_first_question(block_string)
                try:
                    answer = self._get_first_answer(block_string)
                except IndexError:
                    logging.warn(
                        f"Could not find answer in {document.title} for {question}"
                    )
                    continue

                prompts.append(
                    QAFromDoc(
                        question=question,
                        answer=answer,
                        parent_doc=document,
                        line_nr=block_starting_line_nr,
                    )
                )

            block_lines = len(re.findall(r"\n", block_string, flags=re.DOTALL))
            block_starting_line_nr += block_lines

        return prompts



from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass

import genanki
import markdown


@dataclass(frozen=True)
class AnkiPrompt(ABC):
    base_deck: str
    tags: Sequence[str]
    uuid: int  # UUID is a unique identifier for the prompt, used for scheduling. If a new prompt is added with the same uuid, it will be treated as an update to the existing prompt. Otherwise, they will be interpreted as separate prompts.

    def field_to_html(self, field: str) -> str:
        return markdown.markdown(field)

    @property
    @abstractmethod
    def genanki_model(self) -> genanki.Model:
        ...

    @abstractmethod
    def to_genanki_note(self) -> genanki.Note:
        ...

    @property
    def deck(self) -> str:
        deck_prefix = "anki/deck/"
        deck_in_tags = (
            tag.replace(deck_prefix, "").replace("/", "::")
            for tag in self.tags
            if tag.startswith(deck_prefix)
        )
        subdeck = next(deck_in_tags, None)

        if subdeck is None:
            return self.base_deck

        return f"{self.base_deck}::{subdeck}"



from pathlib import Path

from .document_source import MarkdownDocumentSource
from .extractors.extractor_qa import QAPromptExtractor
from .prompt_source import DocumentPromptSource


class TestPromptSource:
    def test_should_deduplicate_prompts(self, tmp_path: Path):
        with (tmp_path / "test.md").open("w") as f:
            f.write(
                """Q. Question
A. Answer

Q. Question
A. Answer"""
            )
        prompts = DocumentPromptSource(
            document_ingester=MarkdownDocumentSource(directory=tmp_path),
            prompt_extractors=[
                QAPromptExtractor(question_prefix="Q.", answer_prefix="A.")
            ],
        ).get_prompts()
        assert len(prompts) == 1



import logging
from pathlib import Path

import pytest

from memium.__main__ import main

from .destination.ankiconnect.ankiconnect_gateway import anki_connect_is_live


@pytest.mark.skipif(
    not anki_connect_is_live(),
    reason="Tests require a running AnkiConnect server",
)
def test_main(
    caplog: pytest.LogCaptureFixture, base_deck: str = "Main Integration Test"
):
    caplog.set_level(logging.INFO)

    # Clear and delete output path
    test_input_path = Path("/output")

    if test_input_path.exists():
        for entity in test_input_path.iterdir():
            if not entity.is_dir():
                entity.unlink()
    else:
        test_input_path.mkdir(parents=True)

    with (test_input_path / "test.md").open("w") as f:
        f.write(
            """Q. Question
A. å
"""
        )

    main(
        base_deck=base_deck,
        input_dir=test_input_path,
        max_deletions_per_run=2,
        dry_run=False,
    )
    assert "Pushing 1 cards to Anki" in caplog.text

    # Test idempotency
    main(
        base_deck=base_deck,
        input_dir=test_input_path,
        max_deletions_per_run=0,  # 0 deletions allowed to test idempotency
        dry_run=False,
    )



from abc import ABC
from collections.abc import Sequence
from dataclasses import dataclass

from ..document import Document
from .prompt import BasePrompt


@dataclass(frozen=True)
class PromptFromDocMixin(BasePrompt, ABC):
    parent_doc: Document
    line_nr: int

    @property
    def scheduling_uid(self) -> int:
        ...

    @property
    def update_uid(self) -> int:
        ...

    @property
    def tags(self) -> Sequence[str]:
        ...



from pathlib import Path

from ..document import Document
from .extractor_qa import QAPromptExtractor


def test_qa_prompt_extractor(tmpdir: Path):
    doc = Document(
        content="""

Q. What is the meaning of life?
A. 42

#anki/tag/test_tag

""",
        source_path=tmpdir / "test.md",
    )

    extractor = QAPromptExtractor(
        question_prefix="Q.", answer_prefix="A."
    ).extract_prompts(doc)

    assert len(extractor) == 1
    prompt = extractor[0]
    assert prompt.question == "What is the meaning of life?"
    assert prompt.answer == "42"
    assert prompt.tags == ["anki/tag/test_tag"]
    assert prompt.scheduling_uid == 3643087944



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Protocol, TypeAlias

from ..source.prompts.prompt import BasePrompt, DestinationPrompt


@dataclass(frozen=True)
class PushPrompts:
    prompts: Sequence[BasePrompt]


@dataclass(frozen=True)
class DeletePrompts:
    prompts: Sequence[DestinationPrompt]


PromptDestinationCommand: TypeAlias = PushPrompts | DeletePrompts


class PromptDestination(Protocol):
    def get_all_prompts(self) -> Sequence[DestinationPrompt]:
        ...

    def update(self, commands: Sequence[PromptDestinationCommand]) -> None:
        ...



import invoke as inv

from memium.tasks.github import get_issues_assigned_to_me, issue_dialog
from memium.tasks.graphite import (
    create_branch_from_issue,
    submit_pr,  # noqa: F401 # type: ignore
)
from memium.tasks.smoketest import (
    smoketest_cli,  # noqa: F401 # type: ignore
    smoketest_docker,  # noqa: F401 # type: ignore
)

PYTEST_CMD = "pytest --durations=5 --cov=memium memium --cov-report xml:.coverage.xml --cov-report lcov:.coverage.lcov"


@inv.task  # type: ignore
def install_dev(c: inv.Context):
    print("--- Installing development dependencies ---")
    c.run("pip install --upgrade .[dev]")
    print("✅✅✅ Development dependencies installed ✅✅✅")


@inv.task  # type: ignore
def install_test(c: inv.Context):
    print("--- Installing development dependencies ---")
    c.run("pip install --upgrade .[dev,tests]")
    print("✅✅✅ Development dependencies installed ✅✅✅")


@inv.task  # type: ignore
def install(c: inv.Context):
    print("--- Installing dependencies ---")
    c.run("pip install --upgrade .")
    print("✅✅✅ Installed package ✅✅✅")


@inv.task  # type: ignore
def generate_coverage(c: inv.Context):
    c.run(PYTEST_CMD)


@inv.task  # type: ignore
def test(c: inv.Context):
    print("--- Testing ---")
    generate_coverage(c)

    coverage_report = c.run("diff-cover .coverage.xml", hide=True)
    if coverage_report is None:
        print("No coverage report generated")
        return

    lines = coverage_report.stdout.split("\n")
    try:
        coverage_line = next(line for line in lines if "Coverage: " in line)
    except StopIteration:
        print("No lines affected by coverage")
        return

    missing_lines = [line for line in lines if "Missing lines" in line]
    coverage_percent = int(coverage_line.split(" ")[1][:-1])

    if coverage_percent < 80:
        print("\n=== Lines with missing coverage ===")
        for line in missing_lines:
            print(line)
        proceed = input(
            f"🚧🚧🚧 Coverage is {coverage_percent}%. Proceed? [y/N] "
        )
        if proceed.lower() != "y":
            print("Aborting")
            return

    print("✅✅✅ Tests passed ✅✅✅")


@inv.task  # type: ignore
def lint(c: inv.Context):
    print("--- Linting ---")
    c.run("ruff format .")
    c.run("ruff . --fix --extend-select F401 --extend-select F841")
    print("✅✅✅ Lint ✅✅✅")


@inv.task  # type: ignore
def types(c: inv.Context):
    print("--- Type-checking ---")
    c.run("pyright memium")
    print("✅✅✅ Types ✅✅✅")


@inv.task(aliases=("next",))  # type: ignore
def branch_from_next_issue(c: inv.Context):
    my_issues = get_issues_assigned_to_me(c)

    if not my_issues:
        print("No issues found")
        return

    selected_issue_index = issue_dialog(my_issues)
    create_branch_from_issue(
        c=c, selected_issue=my_issues[selected_issue_index]
    )


@inv.task  # type: ignore
def validate_ci(c: inv.Context):
    print("--- Validating CI ---")
    lint(c)
    types(c)
    generate_coverage(c)
    print("✅✅✅ CI valid ✅✅✅")


@inv.task(aliases=("new",))  # type: ignore
def new_branch_from_issue(c: inv.Context):
    my_issues = get_issues_assigned_to_me(c)

    if not my_issues:
        print("No issues assigned to you found")
        return

    selected_issue_index = issue_dialog(my_issues)
    c.run("git checkout main")
    c.run("git pull")
    create_branch_from_issue(
        c=c, selected_issue=my_issues[selected_issue_index]
    )



def get_letter_from_alphabet_position(position: int) -> str:
    return chr(
        position + 96
    )  # 96 comes from the Unicode point position for 'a' minus 1


def get_letter_alphabet_position(letter: str) -> int:
    return ord(letter.lower()) - 96



from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Generic, Protocol, TypeVar

from .destination.destination import (
    DeletePrompts,
    PromptDestinationCommand,
    PushPrompts,
)
from .source.prompts.prompt import BasePrompt, DestinationPrompt

K = TypeVar("K")
T = TypeVar("T")
S = TypeVar("S")


class BaseDiffDeterminer(Protocol):
    def sync(
        self,
        source_prompts: Sequence[BasePrompt],
        destination_prompts: Sequence[DestinationPrompt],
    ) -> Sequence[PromptDestinationCommand]:
        ...


@dataclass(frozen=True)
class GeneralSyncer(Generic[K, T, S]):
    source: Mapping[K, T]
    destination: Mapping[K, S]

    def only_in_source(self) -> Sequence[T]:
        return [
            value
            for key, value in self.source.items()
            if key not in self.destination
        ]

    def only_in_destination(self) -> Sequence[S]:
        return [
            value
            for key, value in self.destination.items()
            if key not in self.source
        ]


class PromptDiffDeterminer(BaseDiffDeterminer):
    def sync(
        self,
        source_prompts: Sequence[BasePrompt],
        destination_prompts: Sequence[DestinationPrompt],
    ) -> Sequence[PromptDestinationCommand]:
        # Update prompts if content or tags have changed. This doesn't affect scheduling.
        prompts_to_update = GeneralSyncer(
            source={prompt.update_uid: prompt for prompt in source_prompts},
            destination={
                prompt.prompt.update_uid: prompt
                for prompt in destination_prompts
            },
        ).only_in_source()

        # Only delete prompts whose content have changed. This essentially resets their scheduling.
        prompts_to_delete = GeneralSyncer(
            source={prompt.scheduling_uid: prompt for prompt in source_prompts},
            destination={
                dest_prompt.prompt.scheduling_uid: dest_prompt
                for dest_prompt in destination_prompts
            },
        ).only_in_destination()

        return [
            DeletePrompts(prompts_to_delete),
            PushPrompts(prompts_to_update),
        ]



import hashlib
import re

import unidecode
from bs4 import BeautifulSoup


def replace_whitespace(text: str) -> str:
    return "_".join(text.split())


def remove_non_content_html_tags(text: str) -> str:
    """Remove non-content html tags from a string."""
    soup = BeautifulSoup(text, "html.parser")

    for img in soup.find_all("img"):
        if img.get("src"):
            img.replace_with(img["src"])
        else:
            img.delete()

    return soup.text


def remove_punctuation(text: str) -> str:
    """Clean string before hashing, so changes to spacing, punctuation, newlines etc. do not affect the hash."""
    lowered = text.lower()

    punctuation = r"""!"#$%&'()*+,-./:;<=>?@[\]^_`|~"""
    cleaned = lowered.translate(str.maketrans("", "", punctuation))

    return cleaned


def remove_markdown_links(text: str) -> str:
    """Get the text from markdown links, e.g. '[link text](link url)' -> 'link text'"""
    return re.sub(r"\[([^\]]+)\]\([^\)]+\)", r"\1", text)


def decode_unicode(text: str) -> str:
    """Standardise accents in a string."""
    return unidecode.unidecode(text)


def remove_list_markup(text: str) -> str:
    """Remove markdown list markup, including '1. Item', '*' and '-'."""
    without_list_numbers = re.sub(r"^\s*\d+\. ", "", text, flags=re.MULTILINE)
    without_bullets = re.sub(
        r"^\s\* ", "", without_list_numbers, flags=re.MULTILINE
    )
    without_dashes = re.sub(r"^\s- ", "", without_bullets, flags=re.MULTILINE)
    return without_dashes


def clean_str(input_str: str) -> str:
    """Clean string before hashing, so changes to spacing, punctuation, newlines etc. do not affect the hash."""
    cleaned = input_str

    for cleaner in [
        remove_non_content_html_tags,
        remove_markdown_links,
        remove_list_markup,
        remove_punctuation,
        decode_unicode,
        replace_whitespace,
    ]:
        cleaned = cleaner(cleaned)

    return cleaned


def hash_str_to_int(input_string: str, max_length: int = 10) -> int:
    # Convert the string to bytes
    bytes_string = input_string.encode()

    # Create a SHA256 hash object
    hash_object = hashlib.sha256(bytes_string)

    # Hex digest the hash and convert it to an integer
    unique_int = int(hash_object.hexdigest(), 16)
    shortened = unique_int % 10**max_length

    return shortened



from collections.abc import Sequence
from pathlib import Path

import invoke as inv

from memium.tasks.graphite import submit_pr  # noqa: F401 # type: ignore


def get_code_blocks_from_md(md_path: Path) -> Sequence[str]:
    md = md_path.read_text()
    code_blocks = md.split("```")[1::2]
    return code_blocks


def create_smoketest_dir() -> Path:
    print("💨 Running smoketest")
    input_dir = Path.home() / "smoketest_dir"
    input_dir.mkdir(exist_ok=True)

    test_file = input_dir / "test.md"
    test_file.write_text("Q. Question here\nA. Answer!")
    return input_dir


@inv.task  # type: ignore
def smoketest_docker(c: inv.Context):
    input_dir = create_smoketest_dir()

    code_blocks = get_code_blocks_from_md(Path("readme.md"))
    docker_block = (
        next(block for block in code_blocks if "docker run" in block).strip()
        + " \\\n"
    )

    # Only keep content after docker line
    docker_block = docker_block[docker_block.index("docker run") :]
    replaced_image_location = docker_block.replace(
        "ghcr.io/martinbernstorff/memium:latest", "memium"
    )
    replaced_input_dir = replaced_image_location.replace(
        "$INPUT_DIR", str(input_dir)
    )

    smoketest_docker_command = (
        replaced_input_dir + "  --dry-run \\\n" + "  --skip-sync"
    )

    print(smoketest_docker_command)

    c.run("docker build . -t memium:latest -f Dockerfile")
    c.run("docker volume create ankidecks")
    c.run(smoketest_docker_command)
    print("💨🎉 Smoketest complete")


@inv.task  # type: ignore
def smoketest_cli(c: inv.Context):
    smoketest_dir = create_smoketest_dir()
    code_blocks = get_code_blocks_from_md(Path("readme.md"))
    cli_block = next(block for block in code_blocks if "cli-block" in block)
    sanitised_cli_block = cli_block.splitlines()[1].replace("> ", "")

    cli_smoketest_cmd = sanitised_cli_block.replace(
        "[YOUR_INPUT_DIR]", str(smoketest_dir)
    )
    print(cli_smoketest_cmd)
    c.run(cli_smoketest_cmd + "  --dry-run \\\n" + "  --skip-sync")
    print("💨🎉 Smoketest complete")



from collections.abc import Sequence

import pytest

from .destination.destination import DeletePrompts, PushPrompts
from .diff_determiner import GeneralSyncer, PromptDiffDeterminer
from .source.prompts.prompt import BasePrompt, DestinationPrompt
from .source.prompts.prompt_qa import QAPrompt, QAWithoutDoc


@pytest.fixture()
def diff_determiner() -> PromptDiffDeterminer:
    return PromptDiffDeterminer()


def test_diff_determiner():
    syncer = GeneralSyncer(
        source={"a": 1, "b": 2}, destination={"b": "2", "c": "3"}
    )

    assert syncer.only_in_source() == [1]
    assert syncer.only_in_destination() == ["3"]


from dataclasses import dataclass


@dataclass(frozen=True)
class DiffDeterminerExample:
    example_title: str
    source_prompts: Sequence[BasePrompt]
    destination_prompts: Sequence[DestinationPrompt]
    delete_prompts: Sequence[DestinationPrompt]
    push_prompts: Sequence[BasePrompt]


@pytest.mark.parametrize(
    "example",
    [
        DiffDeterminerExample(
            example_title="Basic",
            source_prompts=[
                QAPrompt(question="a", answer="a"),
                QAPrompt(question="b", answer="b"),
            ],
            destination_prompts=[
                DestinationPrompt(
                    QAPrompt(question="b", answer="b"), destination_id="2"
                ),
                DestinationPrompt(
                    QAPrompt(question="c", answer="c"), destination_id="3"
                ),
            ],
            delete_prompts=[
                DestinationPrompt(
                    QAPrompt(question="c", answer="c"), destination_id="3"
                )
            ],
            push_prompts=[QAPrompt(question="a", answer="a")],
        ),
        (
            DiffDeterminerExample(
                example_title="Updated tags",
                source_prompts=[
                    QAWithoutDoc(question="a", answer="a", add_tags=["NewTag"])
                ],
                destination_prompts=[
                    DestinationPrompt(
                        QAWithoutDoc(
                            question="a", answer="a", add_tags=["OldTag"]
                        ),
                        destination_id="1",
                    )
                ],
                delete_prompts=[],
                push_prompts=[
                    QAWithoutDoc(question="a", answer="a", add_tags=["NewTag"])
                ],
            )
        ),
    ],
)
def test_prompt_diff_determiner(
    diff_determiner: PromptDiffDeterminer, example: DiffDeterminerExample
):
    diff = diff_determiner.sync(
        source_prompts=example.source_prompts,
        destination_prompts=example.destination_prompts,
    )
    assert diff == [
        DeletePrompts(example.delete_prompts),
        PushPrompts(example.push_prompts),
    ]



from collections.abc import Sequence
from pathlib import Path

from .document import Document
from .document_source import BaseDocumentSource
from .extractors.extractor_qa import QAPromptExtractor
from .prompt_source import DocumentPromptSource


def test_document_prompt_source():
    source = DocumentPromptSource(
        document_ingester=FakeDocumentIngester(
            [
                Document(
                    """Q. What is a test even?
A. Nothing""",
                    Path("test.md"),
                )
            ]
        ),
        prompt_extractors=[
            QAPromptExtractor(question_prefix="Q.", answer_prefix="A.")
        ],
    )
    prompts = source.get_prompts()
    assert len(prompts) == 1


class FakeDocumentIngester(BaseDocumentSource):
    def __init__(self, documents: Sequence[Document]):
        self.documents = documents

    def get_documents(self) -> Sequence[Document]:
        return self.documents



from collections.abc import Sequence
from typing import Protocol

from ..document import Document
from ..prompts.prompt import BasePrompt


class BasePromptExtractor(Protocol):
    def extract_prompts(self, document: Document) -> Sequence[BasePrompt]:
        ...
